import numpy as np


def bootcorr(X,
             corr_fn,
             n_draws=30,
             subsample=0.7,
             replace=True,
             random_state=42):
    """Estimate multiple correlation matrices based on bootstrapped samples.

    The produced correlation matrices can be used to check if certain
      correlation estimates are stable.

    Parameters
    ----------
    n_draws : int
        number of bootstraps.

    corr_fn : function
        python function that returns a correlation matrix

    subsample : float
        percentage of samples to use in every estimation.

    replace: bool
        True (default) with replacement, i.e. samples can be used multiple
        times (bootstrapping).
        False is without replacement, i.e. samples are used once, sample are
        not reused (SRSWOR)

    Info:
    -----
    Forked from https://github.com/kmedian/korr/blob/main/korr/bootcorr.py
    but `corr_fn` must return only the correlation matrix without p-values!
    """
    # number of examples
    n_samples, n_features = X.shape

    # size
    if isinstance(subsample, int):
        n_size = subsample
    else:
        n_size = int(n_samples * subsample)

    # create empty tensors
    rho3 = np.zeros(shape=(n_draws, n_features, n_features))
    rho3[:] = np.nan

    oob = []
    rng = set(range(len(X)))

    # set seed
    if random_state:
        np.random.seed(random_state)

    for d in range(n_draws):
        idx = np.random.choice(range(n_samples), size=n_size, replace=replace)
        oob.append(list(rng - set(idx)))
        rho3[d] = corr_fn(X[idx, :])

    # done
    return rho3, oob



import maxjoshua as mh
import numpy as np
import numpy.testing as npt


def test1():
    A = [[1, 1], [0, 0], [1, 1]]
    negate = (0, 0)
    B = [[1, 1], [0, 0], [1, 1]]
    C = mh.negate_bool_features(A, negate)
    npt.assert_array_equal(B, C.astype(int))


def test2():
    A = [[1, 1], [0, 0]]
    negate = (1, 1)
    B = [[0, 0], [1, 1]]
    C = mh.negate_bool_features(A, negate)
    npt.assert_array_equal(B, C.astype(int))



from .bootstrap_solutions import bootstrap_solutions
from .enumerate_negations import enumerate_negations
import numpy as np
from .negate_bool_features import negate_bool_features
from .hard_voting import hard_voting
import korr


def korr_matthews_mcc(x):
    return korr.mcc(x)[0]


def binsel(X: np.array,
           y: np.array,
           n_select: int = 5,
           max_rho: float = 0.4,
           preselect: float = None,
           n_draws: int = 50,
           subsample: float = 0.7,
           replace: bool = False,
           random_state: int = 42,
           unique: bool = True,
           oob_score: bool = True,
           verbose: bool = False) -> (list, list, float, list):
    # bootstrap some possible solutions
    solutions, oob = bootstrap_solutions(
        X, y=y,
        n_select=n_select,
        max_rho=max_rho,
        preselect=preselect,
        n_draws=n_draws,
        subsample=subsample,
        replace=replace,
        random_state=random_state,
        unique=unique,
        verbose=verbose,
        corr_fn=korr_matthews_mcc)

    # find best way to negate features and store evaluation
    results = []
    for i, indicies in enumerate(solutions):
        # convert back to in-sample indicies
        rowidx = list(set(range(len(X))) - set(oob[i]))
        # fit negations based on in-sample
        negate, rho = enumerate_negations(
            X[:, indicies][rowidx, :], y[rowidx])
        # re-evaluate with oob
        if oob_score:
            Xtmp = negate_bool_features(X[:, indicies][oob[i], :], negate)
            Yvote = hard_voting(Xtmp)
            rho = korr.confusion_to_mcc(
                korr.confusion(y[oob[i]], Yvote))
        # save it
        results.append([tuple(indicies), negate, rho])

    # order results
    idx_sorted = np.flip(np.argsort([r[-1] for r in results]))
    results = [results[i] for i in idx_sorted]

    # done
    idx, neg, rho = results[0]
    return idx, neg, rho, results



import maxjoshua as mh
import numpy as np
import numpy.testing as npt


def test1():
    A = np.array([1, 0, 2, 1])
    B = np.array([1, 0, 1, 1])
    C = mh.hard_voting(A)
    npt.assert_array_equal(B, C.astype(int))


def test2():
    A = [1, 0, 2, 1]
    B = np.array([1, 0, 1, 1])
    C = mh.hard_voting(A)
    npt.assert_array_equal(B, C.astype(int))



import numpy as np


def hard_voting(X: np.array) -> np.array:
    # convert to numpy array
    if isinstance(X, (list, tuple)):
        X = np.array(X)

    # return vector directly
    if len(X.shape) == 1:
        return X > 0
    else:
        # majority vote
        n_votes = len(X[0])
        n_majority = int(n_votes / 2.0) + 1
        return np.count_nonzero(X, axis=1) >= n_majority



import numpy as np
import korr
from .bootcorr import bootcorr


def bootstrap_solutions_all(X: np.array,
                            n_select: int = 5,
                            max_rho: float = 0.4,
                            n_draws: int = 50,
                            subsample: float = 0.7,
                            replace: bool = False,
                            random_state: int = 42,
                            unique: bool = True,
                            verbose: bool = False,
                            corr_fn=None) -> (np.array, list):
    # compute all Matthew's correlations between (Xi,Xj)
    rho3, oob = bootcorr(
        X,
        corr_fn=corr_fn,
        n_draws=n_draws,
        subsample=subsample,
        replace=replace,
        random_state=random_state)

    # for each draw find the lowest abs(corr(Xi,Xj))
    solutions = []   # store results in list
    for cmat in rho3:
        idxlo = korr.mincorr(
            cmat, n_stop=n_select, max_rho=max_rho,
            verbose=verbose)
        solutions.append(sorted(idxlo))

    # done
    if unique:
        sol, idx = np.unique(solutions, axis=0, return_index=True)
        return sol, list(np.array(oob)[idx])
    else:
        return solutions, oob


def bootstrap_solutions_pre(X: np.array,
                            y: np.array,
                            n_select: int = 5,
                            max_rho: float = 0.4,
                            preselect: float = 0.8,
                            n_draws: int = 50,
                            subsample: float = 0.7,
                            replace: bool = False,
                            random_state: int = 42,
                            unique: bool = True,
                            verbose: bool = False,
                            corr_fn=None) -> (np.array, list):
    # convert to number of features to preselect
    n_features = len(X[0])
    if isinstance(preselect, float):
        n_pre = max(1, int(n_features * min(1.0, preselect)))
    elif isinstance(preselect, int):
        n_pre = max(1, min(n_features, preselect))
    else:
        n_pre = None  # all

    # compute all Matthew's correlations between (y,X) and (Xi,Xj)
    rho3, oob = bootcorr(
        np.c_[y, X],
        corr_fn=corr_fn,
        n_draws=n_draws,
        subsample=subsample,
        replace=replace,
        random_state=random_state)

    # for each draw find the lowest abs(corr(Xi,Xj))
    solutions = []   # store results in list
    for cmat in rho3:
        # a) sort abs corr(y,x), b) sort by biggest, c) add +1 to index bc y
        idxft = np.flip(np.argsort(np.abs(cmat[0, 1:]))) + 1
        idxft = idxft[:n_pre]
        # find the smallest corr(xi,xj) between selected features
        idxlo = korr.mincorr(
            cmat[:, idxft][idxft, :],
            n_stop=n_select,
            max_rho=max_rho,
            verbose=verbose)
        # convert indicies back to 'X' matrix (not 'dat')
        idxlo = list(np.array(idxft[idxlo]) - 1)
        # save it
        solutions.append(sorted(idxlo))

    # done
    if unique:
        sol, idx = np.unique(solutions, axis=0, return_index=True)
        return sol, list(np.array(oob)[idx])
    else:
        return solutions, oob


def bootstrap_solutions(X: np.array,
                        y: np.array = None,
                        n_select: int = 5,
                        max_rho: float = 0.4,
                        preselect: float = None,
                        n_draws: int = 50,
                        subsample: float = 0.7,
                        replace: bool = False,
                        random_state: int = 42,
                        unique: bool = True,
                        verbose: bool = False,
                        corr_fn=None) -> (np.array, list):
    if preselect and y is not None:
        return bootstrap_solutions_pre(
            X, y, n_select=n_select, max_rho=max_rho, preselect=preselect,
            n_draws=n_draws, subsample=subsample, replace=replace,
            random_state=random_state, unique=unique, verbose=verbose,
            corr_fn=corr_fn)
    else:
        return bootstrap_solutions_all(
            X, n_select=n_select, max_rho=max_rho,
            n_draws=n_draws, subsample=subsample, replace=replace,
            random_state=random_state, unique=unique, verbose=verbose,
            corr_fn=corr_fn)



__version__ = '0.4.3'

from .negate_bool_features import negate_bool_features
from .hard_voting import hard_voting
from .enumerate_negations import enumerate_negations
from .bootstrap_solutions import (
    bootstrap_solutions_all, bootstrap_solutions_pre, bootstrap_solutions)
from .binsel import binsel
from .fltsel import fltsel
from .bootcorr import bootcorr
from .sparsenn import (
    pretrain_submodels, SparseLayerAsEnsemble, InverseTransformer)



import maxjoshua as mh
import korr
import numpy as np


def korr_matthews_mcc(x):
    return korr.mcc(x)[0]


def test1():
    X = np.random.random((1000, 21)) > 0.5
    solutions, oob = mh.bootstrap_solutions_pre(
        X=X[:, :20], y=X[:, -1], corr_fn=korr_matthews_mcc,
        n_select=5, preselect=1.0, n_draws=20, subsample=0.3)
    assert len(solutions) == 20


def test2():
    X = np.random.random((1000, 21)) > 0.5
    solutions, oob = mh.bootstrap_solutions_pre(
        X=X[:, :20], y=X[:, -1], corr_fn=korr_matthews_mcc,
        n_select=5, preselect=0.6, n_draws=20, subsample=0.3)
    assert len(solutions) == 20


def test3():
    X = np.random.random((1000, 21)) > 0.5
    solutions, oob = mh.bootstrap_solutions_all(
        X, corr_fn=korr_matthews_mcc,
        n_select=5, n_draws=20, subsample=0.3)
    assert len(solutions) == 20



import maxjoshua as mh
import numpy as np


def test1():
    def corr_fn(x):
        return np.corrcoef(x, rowvar=False)
    X = np.random.random((100, 3))
    rho3, _ = mh.bootcorr(X, corr_fn=corr_fn, n_draws=10)
    assert rho3.shape == (10, 3, 3)



from .fltsel import fltsel
import numpy as np
import itertools
import gc
import tensorflow as tf
from keras_tweaks import dense_sparse_matmul
from typing import List


def pretrain_submodels(X, y,
                       num_out: int = 64,
                       n_select: int = 5,
                       n_draws: int = 100):
    """ Forward selection of N submodels and create params for tf COO tensor

    Example:
    --------
    indices, values, num_in, num_out = pretrain_submodels(
        X, y, num_out=64, n_select=3)
    W = tf.sparse.SparseTensor(
        indices=indices, values=values, dense_shape=(num_in, num_out))
    """
    # increase draws
    n_draws_ = max(max(X.shape[1] * 2, num_out * 2), n_draws)

    # adjust `y`
    if len(y.shape) == 1:
        y = y.reshape(-1, 1)

    # run mh.fltsel for each target variable
    results = []
    for i in range(y.shape[1]):
        idx, loss, beta, res = fltsel(
            X, y[:, i], preselect=0.8, oob_score=True, subsample=0.5,
            n_select=n_select, unique=True, n_draws=n_draws_)
        print(i, idx, loss, beta)
        results.extend(res)

    # filter unique; select submodel with smallest loss
    ures = {}
    for res in results:
        tmp = ures.get(res[0], {})
        if len(tmp) == 0:
            ures[res[0]] = {"beta": res[1], "loss": res[2]}
        else:
            prevloss = tmp.get("loss", np.inf)
            if prevloss > res[2]:
                ures[res[0]] = {"beta": res[1], "loss": res[2]}

    del results
    gc.collect()

    # convert back
    results = [(r[0], r[1].get('beta'), r[1].get('loss'))
               for r in ures.items()]
    results.sort(key=lambda x: x[-1])
    del ures
    gc.collect()

    # find unused indicies
    allidx = set(range(X.shape[1]))
    selectedidx = set(itertools.chain(*[r[0] for r in results]))
    unused = list(allidx.difference(selectedidx))

    if len(unused) > 0:
        pct_unused = len(unused) / X.shape[1] * 100
        print(f"Unused features: {len(unused)} ({pct_unused:.1f}%)")

    if len(results) < num_out:
        print(f"Insufficent submodels identified: {len(results)} of {num_out}")

    # select the best submodels
    results = results[:num_out]

    # convert to COO indices and values
    indices = []
    values = []
    for j, res in enumerate(results):
        indices.extend([(i, j) for i in res[0]])
        values.extend(res[1])

    # done
    num_in = X.shape[1]
    return indices, values, num_in, num_out


class SparseLayerAsEnsemble(tf.keras.layers.Layer):
    def __init__(self,
                 num_in: int,
                 num_out: int,
                 sp_indices: List[List[int]],
                 sp_values: List[float],
                 sp_trainable: bool = False,
                 norm_trainable: bool = True,
                 **kwargs):
        super(SparseLayerAsEnsemble, self).__init__(**kwargs)
        # batch norm
        self.norm_trainable = norm_trainable
        if self.norm_trainable:
            self.norm = tf.keras.layers.BatchNormalization(
                center=True, scale=True, trainable=True,
                name="normalize_inputs")
        # sparse tensor
        self.num_in = num_in
        self.num_out = num_out
        self.sp_indices = sp_indices
        self.sp_weights = tf.Variable(
            initial_value=sp_values,
            trainable=sp_trainable,
            name='sparse_weights')

    def _get_sp(self):
        return tf.sparse.SparseTensor(
            dense_shape=(self.num_in, self.num_out),
            indices=self.sp_indices,
            values=self.sp_weights)

    def call(self, inputs: tf.Tensor):
        if self.norm_trainable:
            h = self.norm(inputs)
        else:
            h = inputs
        W = self._get_sp()
        h = dense_sparse_matmul(h, W)
        return h


class InverseTransformer(tf.keras.layers.Layer):
    """ Train the inverse transform """
    def __init__(self, units, init_bias=0., init_scale=1., **kwargs):
        super(InverseTransformer, self).__init__(**kwargs)
        self.units = units
        self.init_bias = init_bias
        self.init_scale = init_scale
        self.scale = tf.Variable(
            initial_value=tf.ones(self.units) * self.init_scale,
            trainable=True, name='scale')
        self.bias = tf.Variable(
            initial_value=tf.zeros(self.units) + self.init_bias,
            trainable=True, name='bias')
        self.ops = tf.keras.layers.Lambda(
            lambda x: x * tf.maximum(self.scale, 1e-8) + self.bias)

    def call(self, inputs):
        return self.ops(inputs)

    def get_config(self):
        config = super().get_config().copy()
        config.update({
            'units': self.units,
            'init_bias': self.init_bias,
            'init_scale': self.init_scale})
        return config



from itertools import product
from .negate_bool_features import negate_bool_features
from .hard_voting import hard_voting
from korr import confusion, confusion_to_mcc
import numpy as np


def enumerate_negations(X: np.array, y: np.array) -> (list, float):
    n_features = len(X[0])  # warning 15 => 32768 combinations!

    best_neg = None
    best_rho = 0.0
    for negate in product([0, 1], repeat=n_features):
        Xtmp = negate_bool_features(X, negate)
        Yvote = hard_voting(Xtmp)
        rho = confusion_to_mcc(confusion(y, Yvote))
        if rho > best_rho:
            best_neg = negate
            best_rho = rho

    return best_neg, best_rho



import maxjoshua as mh
import numpy as np


def test1():
    X = np.random.random((1000, 21)) > 0.5
    idx, neg, rho, results = mh.binsel(
        X=X[:, :20], y=X[:, -1], n_select=5, oob_score=True)
    idx, neg, rho
    assert len(idx) == 5
    assert len(neg) == 5
    assert isinstance(rho, float)



from .bootstrap_solutions import bootstrap_solutions
import numpy as np
import numba
import numpy_linreg.ridge
import numpy_linreg.metrics


@numba.njit
def numba_corrcoef_pearson(x):
    return np.corrcoef(x, rowvar=False)


def fltsel(X: np.array,
           y: np.array,
           n_select: int = 5,
           max_rho: float = 0.4,
           preselect: float = None,
           n_draws: int = 50,
           subsample: float = 0.7,
           replace: bool = False,
           random_state: int = 42,
           unique: bool = True,
           oob_score: bool = True,
           verbose: bool = False,
           l2: float = 0.01) -> (list, list, float, list):
    # bootstrap some possible solutions
    solutions, oob = bootstrap_solutions(
        X, y=y,
        n_select=n_select,
        max_rho=max_rho,
        preselect=preselect,
        n_draws=n_draws,
        subsample=subsample,
        replace=replace,
        random_state=random_state,
        unique=unique,
        verbose=verbose,
        corr_fn=numba_corrcoef_pearson)

    # find best way to negate features and store evaluation
    results = []
    for i, indicies in enumerate(solutions):
        feat_cols = solutions[i]
        oob_rows = oob[i]
        # fit model
        beta = numpy_linreg.ridge.lu(
            y=y[oob_rows], X=X[:, feat_cols][oob_rows, :], lam=l2)
        # compute RMSE loss
        loss = numpy_linreg.metrics.rmse(
            y=y[oob_rows], X=X[:, feat_cols][oob_rows, :], beta=beta)
        # save it
        results.append([tuple(indicies), beta.tolist(), loss])

    # order results
    idx_sorted = np.argsort([r[-1] for r in results])
    results = [results[i] for i in idx_sorted]

    # done
    idx, beta, loss = results[0]
    return idx, beta, loss, results
